@@ -9,6 +9,7 @@ from macrostrat.utils import get_logger
 from sqlalchemy.ext.compiler import compiles
 from sqlalchemy.sql.expression import Insert
 
+
 from .utils import (
     run_sql,
     get_or_create,
@@ -30,7 +31,7 @@ class Database(object):
     session: Session
     __inspector__ = None
 
-    def __init__(self, db_conn, app=None, echo_sql=False, **kwargs):
+    def __init__(self, db_conn, echo_sql=False, **kwargs):
         """
         We can pass a connection string, a **Flask** application object
         with the appropriate configuration, or nothing, in which
@@ -41,9 +42,7 @@ class Database(object):
         compiles(Insert, "postgresql")(prefix_inserts)
 
         log.info(f"Setting up database connection '{db_conn}'")
-        self.engine = create_engine(
-            db_conn, executemany_mode="batch", echo=echo_sql, **kwargs
-        )
+        self.engine = create_engine(db_conn, echo=echo_sql, **kwargs)
         self.metadata = kwargs.get("metadata", metadata)
 
         # Scoped session for database
