@@ -1,9 +1,9 @@
 from click import echo, secho
 from sqlalchemy.exc import ProgrammingError, IntegrityError, InternalError
 from sqlparse import split, format
-from sqlalchemy.sql import ClauseElement
-from sqlalchemy.orm import sessionmaker, Session
-from sqlalchemy.engine import Engine, Connection, Transaction
+from sqlalchemy.sql.elements import TextClause, ClauseElement
+from sqlalchemy.orm import sessionmaker
+from sqlalchemy.engine import Engine, Connection
 from sqlalchemy.schema import Table
 from sqlalchemy import MetaData, create_engine, text
 from contextlib import contextmanager
@@ -16,6 +16,8 @@ from pathlib import Path
 from warnings import warn
 from psycopg2.sql import SQL, Composable, Composed
 from re import search
+from macrostrat.utils import get_logger
+from .postgresql import _setup_psycopg2_wait_callback
 
 log = get_logger(__name__)
 
@@ -84,15 +86,7 @@ def get_dataframe(connectable, filename_or_query, **kwargs):
 
 def pretty_print(sql, **kwargs):
     for line in sql.split("\n"):
-        for i in [
-            "SELECT",
-            "INSERT",
-            "UPDATE",
-            "CREATE",
-            "DROP",
-            "DELETE",
-            "ALTER",
-        ]:
+        for i in ["SELECT", "INSERT", "UPDATE", "CREATE", "DROP", "DELETE", "ALTER"]:
             if not line.startswith(i):
                 continue
             start = line.split("(")[0].strip().rstrip(";").replace(" AS", "")
@@ -113,12 +107,15 @@ def get_sql_text(sql, interpret_as_file=None, echo_file_name=True):
 
     return sql
 
+
 def _get_queries(sql, interpret_as_file=None):
     if isinstance(sql, (list, tuple)):
         queries = []
         for i in sql:
             queries.extend(_get_queries(i, interpret_as_file=interpret_as_file))
         return queries
+    if isinstance(sql, TextClause):
+        return [sql]
     if isinstance(sql, SQL):
         return [sql]
 
@@ -134,11 +131,14 @@ def _get_queries(sql, interpret_as_file=None):
 
     return split(sql)
 
+
 def _is_prebind_param(param):
     return isinstance(param, Composable)
 
+
 def _split_params(params):
-    params = params or []
+    if params is None:
+        return None, None
     new_params = []
     new_bind_params = []
     if isinstance(params, (list, tuple)):
@@ -159,6 +159,7 @@ def _split_params(params):
         new_bind_params = None
     return new_params, new_bind_params
 
+
 def _get_cursor(connectable):
     if isinstance(connectable, Engine):
         conn = connectable.connect()
@@ -167,17 +168,20 @@ def _get_cursor(connectable):
     conn = connectable
     if hasattr(conn, "raw_connection"):
         conn = conn.raw_connection()
-    while hasattr(conn, "connection"):
-        if callable(conn.connection):
-            conn = conn.connection()
+    while hasattr(conn, "driver_connection") or hasattr(conn, "connection"):
+        if hasattr(conn, "driver_connection"):
+            conn = conn.driver_connection
         else:
             conn = conn.connection
+        if callable(conn):
+            conn = conn()
     if hasattr(conn, "cursor"):
         conn = conn.cursor()
 
     return conn
 
-def _get_connection(connectable):
+
+def _get_connection(connectable) -> Connection:
     if isinstance(connectable, Engine):
         return connectable.connect()
     if isinstance(connectable, Connection):
@@ -188,7 +192,7 @@ def _get_connection(connectable):
     if callable(conn):
         return conn()
     return conn
-    
+
 
 def _render_query(query: Union[SQL, Composed], connectable: Union[Engine, Connection]):
     """Render a query to a SQL string."""
@@ -198,16 +202,29 @@ def _render_query(query: Union[SQL, Composed], connectable: Union[Engine, Connec
     conn = _get_cursor(connectable)
     return query.as_string(conn)
 
-            
+
+def infer_has_server_binds(sql):
+    return "%s" in sql or search(r"%\(\w+\)s", sql)
+
+
 def _run_sql(connectable, sql, **kwargs):
+    """
+    Internal function for running a query on a SQLAlchemy connectable,
+    which always returns an iterator. The wrapper function adds the option
+    to return a list of results.
+    """
     if isinstance(connectable, Engine):
         with connectable.connect() as conn:
             yield from _run_sql(conn, sql, **kwargs)
             return
 
+    _setup_psycopg2_wait_callback()
+
     params = kwargs.pop("params", None)
     stop_on_error = kwargs.pop("stop_on_error", False)
     raise_errors = kwargs.pop("raise_errors", False)
+    has_server_binds = kwargs.pop("has_server_binds", None)
+
     if stop_on_error:
         raise_errors = True
         warn(DeprecationWarning("stop_on_error is deprecated, use raise_errors"))
@@ -241,22 +258,24 @@ def _run_sql(connectable, sql, **kwargs):
             if isinstance(query, (SQL, Composed)):
                 query = _render_query(query, connectable)
 
-            sql_text = query
-            has_server_binds = False
+            sql_text = str(query)
             if isinstance(query, str):
                 sql_text = format(query, strip_comments=True).strip()
                 if sql_text == "":
                     continue
                 # Check for server-bound parameters in sql native style. If there are none, use
                 # the SQLAlchemy text() function, otherwise use the raw query string
-                has_server_binds = "%s" in sql_text or search(r'%\(\w+\)s', sql_text)
+                if has_server_binds is None:
+                    has_server_binds = infer_has_server_binds(sql_text)
 
             log.debug("Executing SQL: \n %s", query)
             if has_server_binds:
                 conn = _get_connection(connectable)
                 res = conn.exec_driver_sql(query, params)
             else:
-                res = connectable.execute(text(query), params=params)
+                if not isinstance(query, TextClause):
+                    query = text(query)
+                res = connectable.execute(query, params)
             yield res
             if trans is not None:
                 trans.commit()
@@ -284,6 +303,33 @@ def run_sql_file(connectable, filename, **kwargs):
 
 
 def run_sql(*args, **kwargs):
+    """
+    Run a query on a SQLAlchemy connectable.
+
+    Parameters
+    ----------
+    connectable : Union[Engine, Connection]
+        A SQLAlchemy engine or connection object.
+    sql : Union[str, Path, IO, SQL, Composed]
+        A SQL query, or a file containing a SQL query.
+    params : Union[dict, list, tuple]
+        Parameters to bind to the query. If a list or tuple, the parameters
+        will be bound to the query in order. If a dict, the parameters will
+        be bound to the query by name.
+    stop_on_error : bool
+        If True, stop running queries if an error is encountered.
+    raise_errors : bool
+        If True, raise errors encountered while running queries.
+    has_server_binds : bool
+        Interpret the query to have server-side bind parameters (requiring execution
+        with the backend driver). By default, this is inferred from the query string,
+        but inference is not always reliable.
+    interpret_as_file : bool
+        If True, force interpreting the query as a file path.
+    yield_results : bool
+        If True, yield the results of the query as they are executed, rather than
+        returning a list after completion.
+    """
     res = _run_sql(*args, **kwargs)
     if kwargs.pop("yield_results", False):
         return res
@@ -401,11 +447,4 @@ def reflect_table(engine, tablename, *column_args, **kwargs):
     """
     schema = kwargs.pop("schema", "public")
     meta = MetaData(schema=schema)
-    return Table(
-        tablename,
-        meta,
-        *column_args,
-        autoload=True,
-        autoload_with=engine,
-        **kwargs,
-    )
+    return Table(tablename, meta, *column_args, autoload_with=engine, **kwargs)
